from __future__ import annotations

import argparse
import os
import re
import textwrap
from dataclasses import dataclass
from enum import Enum
from typing import List

import TexSoup

from .helpers import SupportedLanguages, error_descriptions


class LatCat(Enum):
    STR = 0
    ENV = 1
    CMD = 2
    BRACES = 3
    BRACKETS = 4


@dataclass
class Treenode:
    cat: LatCat = LatCat.STR
    token: str = None
    args: List[Treenode] = None
    children: List[Treenode] = None
    prev_sibling: Treenode = None
    next_sibling: Treenode = None
    prev_node: Treenode = None
    next_node: Treenode = None
    parent: Treenode = None
    is_in_math: bool = False
    is_in_display_math: bool = False
    is_in_arg: bool = False
    pos: int = None
    tex: str = None


AGG_STRINGS = True


def is_command_with_no_text_semantics(node: Treenode):
    if node is None:
        return False
    return node.cat == LatCat.CMD and node.token in ['label', 'ref', 'eqref', 'tag']


def is_display_math(x):
    if x is None:
        return False
    display_envs = ['$$', 'align', 'align*', 'eqnarray', 'eqnarray*', 'equation', 'equation*', 'gather', 'gather*',
                    'multline', 'multline*', 'displaymath', 'flalign', 'flalign*', 'alignat', 'alignat*']
    if isinstance(x, Treenode):
        return (x.cat == LatCat.ENV) and (x.token in display_envs)
    return x in display_envs


def is_numbered_display_math(x):
    if x is None:
        return False
    numbered_display_envs = ['align', 'eqnarray', 'equation', 'gather', 'multline', 'flalign', 'alignat']
    if isinstance(x, Treenode):
        return (x.cat == LatCat.ENV) and (x.token in numbered_display_envs)
    return x in numbered_display_envs


def is_math(x):
    if x is None:
        return False
    if isinstance(x, Treenode):
        return (x.token == 'math') or is_display_math(x.token)
    return (x == 'math') or is_display_math(x)


def initiates_text_mode(x):
    if x is None:
        return False
    text_envs = ['mbox', 'hbox', 'text', 'textnormal', 'textrm', 'textit', 'textsf', 'texttt']
    if isinstance(x, Treenode):
        return x.token in text_envs
    return x in text_envs


def has_descendant(node, predicate, recursive_call=False):
    if node is None:
        return False
    if predicate(node) and recursive_call:
        return True
    for nodelist in [node.args, node.children]:
        if nodelist is not None:
            for child in nodelist:
                if has_descendant(child, predicate, recursive_call=True):
                    return True
    return False


def has_child(node, predicate):
    if node is None:
        return False
    for nodelist in [node.args, node.children]:
        if nodelist is not None:
            for child in nodelist:
                if predicate(child):
                    return True
    return False


def has_sibling(node, predicate):
    if node is None or node.parent is None:
        return False
    return has_child(node.parent, (lambda x: x is not node and predicate(x)))


def is_pure_argument(node):
    while node.parent and node.parent.cat in [LatCat.BRACES, LatCat.BRACKETS]:
        node = node.parent
    return is_command_with_no_text_semantics(node)


def traverse_tex(tex, parent_node_list=None, depth=0):
    global AGG_STRINGS
    no_parent = False
    if parent_node_list is None:
        parent_node_list = []
        no_parent = True
    if isinstance(tex, str):
        if tex.startswith('%'):
            return
        if AGG_STRINGS and len(parent_node_list) > 0 and parent_node_list[-1].cat == LatCat.STR:
            parent_node_list[-1].token += tex
        else:
            parent_node_list.append(Treenode(
                cat=LatCat.STR,
                token=tex
            ))
        return

    new_node = Treenode(pos=tex.position, tex=str(tex))

    s = str(tex)

    if s.startswith(rf'\begin{{{tex.name}}}') or tex.name == '$$':
        new_node.cat = LatCat.ENV
        new_node.token = str(tex.name)
    elif tex.name == '$' or tex.name == 'math':
        new_node.cat = LatCat.ENV
        new_node.token = 'math'
    elif tex.name == 'displaymath':
        new_node.cat = LatCat.ENV
        new_node.token = 'displaymath'
    elif tex.name == '[tex]':
        new_node.cat = LatCat.ENV
        new_node.token = 'document'
    elif s.startswith('{'):
        new_node.cat = LatCat.BRACES
    elif s.startswith('['):
        new_node.cat = LatCat.BRACKETS
    else:
        new_node.cat = LatCat.CMD
        new_node.token = str(tex.name)

    if tex.args:
        new_node.args = []
        for t in tex.args:
            traverse_tex(t, new_node.args, depth + 1)

    new_node.children = None
    if not is_command_with_no_text_semantics(new_node):
        new_node.children = []
        for t in tex.contents:
            traverse_tex(t, new_node.children, depth + 1)
        if new_node.args and not AGG_STRINGS:
            n_arg_tokens = sum(len(node.children) for node in new_node.args)
            new_node.children = new_node.children[n_arg_tokens:]

    for nodelist in [new_node.args, new_node.children]:
        if nodelist is None:
            continue
        prev_sibling = None
        for child in nodelist:
            if prev_sibling:
                prev_sibling.next_sibling = child
            child.prev_sibling = prev_sibling
            child.parent = new_node
            prev_sibling = child

    parent_node_list.append(new_node)
    if no_parent:
        return new_node


def fill_node_links(root_node):
    all_nodes = list()

    def recursive_fill_all_nodes_list(node, math_mode=0, in_args=False):
        nonlocal all_nodes
        if in_args:
            node.is_in_arg = True
        if math_mode:
            node.is_in_math = True
        if math_mode == 2:
            node.is_in_display_math = True
        if is_math(node) or node.is_in_math and not initiates_text_mode(node):
            math_mode = 1
        elif (node.is_in_display_math or is_display_math(node)) and not initiates_text_mode(node):
            math_mode = 2
        all_nodes.append(node)
        for nodelist in [node.args, node.children]:
            if nodelist is not None:
                for child in nodelist:
                    recursive_fill_all_nodes_list(child, math_mode, in_args or (nodelist is node.args) and is_command_with_no_text_semantics(node))

    recursive_fill_all_nodes_list(root_node)
    for i in range(1, len(all_nodes)):
        all_nodes[i].prev_node = all_nodes[i - 1]
        all_nodes[i - 1].next_node = all_nodes[i]
    return all_nodes


def find_prev(node, condition):
    while node is not None:
        node = node.prev_node
        if (node is not None) and condition(node):
            return node


def find_next(node, condition):
    while node is not None:
        node = node.next_node
        if (node is not None) and condition(node):
            return node


def find_child(node, condition):
    if node is None:
        return None
    for nodelist in [node.args, node.children]:
        if nodelist is not None:
            for child in nodelist:
                if condition(child):
                    return child
    return None


def find_descendant(node, condition):
    if node is None:
        return None
    if condition(node):
        return node
    for nodelist in [node.args, node.children]:
        if nodelist is not None:
            for child in nodelist:
                d = find_descendant(child, condition)
                if d is not None:
                    return d
    return None


def tree_to_str(node, depth=0):
    s = ''
    s += f"{'  ' * depth} {node.cat} {repr(node.token or '')}\n"
    for nodelist in [node.args, node.children]:
        if nodelist is not None and nodelist != []:
            if nodelist is node.args:
                s += f"{'  ' * depth} ARGS:\n"
            else:
                s += f"{'  ' * depth} CHILDREN:\n"
            for child in nodelist:
                s += tree_to_str(child, depth + 1)
    return s


def perform_checks(source: str, language: SupportedLanguages = SupportedLanguages.EN, debug_mode=False):
    errors = {}

    def add_error(err_code, bad_node=None):
        nonlocal errors
        if err_code not in error_descriptions[language]:
            return
        pos = None
        if isinstance(bad_node, int):
            pos = bad_node
        elif isinstance(bad_node, Treenode):
            pos = 0
            while bad_node is not None and bad_node.pos is None:
                bad_node = bad_node.prev_node
            if bad_node and bad_node.pos and bad_node.pos:
                pos = bad_node.pos

        if err_code not in errors:
            errors[err_code] = []
        if pos is not None:
            errors[err_code].append(pos)

    try:
        soup = TexSoup.TexSoup(source)
    except Exception as e:
        e = str(e)
        offset = re.search(r'\[Line:? \d+, Offset:? (\d+)]', e)
        if offset:
            offset = int(offset.group(1))
        add_error('PARSE_ERROR', offset)
        return errors

    try:
        latex_tree = traverse_tex(soup)
        all_nodes = fill_node_links(latex_tree)
    except Exception as _:
        add_error('PARSE_ERROR', None)
        return errors

    if debug_mode:
        print(tree_to_str(latex_tree))

    node = find_descendant(latex_tree, lambda n: n.token == '$$')
    if node is not None:
        add_error('DOUBLE_DOLLARS', node.pos)

    re_td = re.compile(r"т\. ?(д\.|н\.|ч\.|к\.)", re.IGNORECASE)
    re_dash_no_spaces = re.compile(r"--([^- ~\n]|$)|(^|[^- ~\n])--")
    re_dash_as_hyphen = re.compile(r"(^|\s)-\s+|\s+-(\s|$)")
    re_multiplication_star = re.compile(r"[^_^]\*|.\*")
    re_space_before_punctuation = re.compile(r"\s+[?!.,;:]")
    re_space_after_punctuation = re.compile(r"[?!.,;:][^ ~\t\n\\]")
    re_space_before_parenthesis = re.compile(r"[^()\[\]{}\n\t-/+]\(")
    re_space_after_parenthesis = re.compile(r"\(\s")
    re_cyrillic_tricky_letter = re.compile(r"[уехаос]")
    re_latin_c_in_rus_text = re.compile(r"[а-яё]\s*c|c\s*[а-яё]", re.IGNORECASE)
    re_math_command = re.compile(r"(\\(infty|cdot|sum))|([0-9 \n]+ *[=+*^])|([+*^] *[0-9 \n]+)")
    re_latin_letter_outside_math_ru = re.compile(r" (^|[, .~])[a-zA-Z]($|[,.:!? ~-]) ")
    re_latin_letter_outside_math_en = re.compile(r" (^|[, .~])[b-zA-HJ-Z]($|[,.:!? ~-]) ")
    re_capitalization_after_comma = re.compile(r"[,;:]\s*[А-ЯЁA-Z]")
    re_capitalization_after_period = re.compile(r"\.\s*[а-яёa-z]")
    re_starts_with_lowercase = re.compile(r"^\s*[а-яёa-z].*", re.DOTALL)
    re_starts_with_uppercase = re.compile(r"^\s*[А-ЯЁA-Z].*", re.DOTALL)
    re_ends_with_space = re.compile(r".*[ \n\t]$", re.DOTALL)
    re_starts_with_cyrillic = re.compile(r"^[а-яё]", re.IGNORECASE)
    re_ends_with_digit_or_letter = re.compile(r".*[0-9a-z]\s*$", re.IGNORECASE | re.DOTALL)
    re_possibly_word = re.compile(r"([^a-z\\]|^)([a-z]{4,}|bad|[a-z]{2,3}\.)", re.IGNORECASE)
    re_multiple_spaces = re.compile(r"(~|\\:|\\ |\\,|\\!|\\>|\\space|\{ }){2,}")
    re_trivial_label = re.compile(r"\{?\s*(eq|equation|eqn|th|thm|lemma|theorem|lem|fig|figure)?:?[^a-z}]\s*}?", re.IGNORECASE)
    re_nonsymbolic_reference = re.compile(r"(рис(унок|унка|унке|\.)|формул(а|е|ой|у|ы)|(равенств|тождеств)(о|а|е|у|ами|ах)|(соотношени|выражени)(е|ю|и|я|ями|ях|ям))\s+\(?\d+\)?", re.IGNORECASE)
    re_mod_cmd = re.compile(r"\bmod\b")
    re_math_no_backslash = re.compile(r"([^\\a-z]|^)(cos|csc|exp|ker|limsup|max|min|sinh|arcsin|cosh|deg|gcd|lg|ln|Pr|sup|arctan|cot|det|hom|lim|log|sec|tan|arg|coth|dim|liminf|sin|tanh)[^a-z]")
    re_ru_ordinal = re.compile(r"\s*-{1,2}\s*(ый|ого|о|тому|ому|ему|ом|ая|ой|ую|ые|ыми|и|ым|тым|той|им|его|того|тых|ых|том|ем|ём|ех|ёх|ух)([^а-яё]|$)", re.IGNORECASE | re.DOTALL | re.UNICODE)
    re_small_numeral = re.compile(r"([,.!?:]|\W\s+)[0-5]([,.!?:]|\s+\W)")

    for node in all_nodes:
        if node.cat == LatCat.STR:
            if re_multiplication_star.search(node.token):
                add_error('MULTIPLICATION_SIGN', node)
            if re_multiple_spaces.search(node.token):
                add_error('INDENTATION_WITH_SPACES', node)
            if re_starts_with_cyrillic.search(node.token) and node.prev_node.cat == LatCat.CMD:
                add_error('NO_SPACE_AFTER_COMMAND_BEFORE_CYRILLIC', node)

        if is_numbered_display_math(node) and not has_descendant(node, (lambda x: x.token == 'label')):
            add_error('NUMBERED_MATH_NEEDS_REFERENCING', node)
        if is_display_math(node):
            if node.next_sibling:
                if is_display_math(node.next_sibling):
                    add_error('CONSECUTIVE_DISPLAY_FORMULAE', node.next_sibling.pos or node.pos)
                if (node.next_sibling.cat == LatCat.STR) and node.next_sibling.token.startswith('\\\\'):
                    add_error('LINEBREAK_AFTER_DISPLAY_FORMULAE', node.next_sibling.pos or node.pos)
            if node.prev_sibling and (node.prev_sibling.cat == LatCat.STR) and node.prev_sibling.token.startswith(
                    '\\\\'):
                add_error('LINEBREAK_BEFORE_DISPLAY_FORMULAE', node.pos or node.prev_sibling.pos)

        if node.token in ['eqnarray', 'eqnarray*']:
            add_error('EQNARRAY_USED', node.pos)

        if is_math(node) and len(node.children) == 1 and node.children[0].cat != LatCat.STR:
            add_error('UNNECESSARY_MATH_MODE', node)

        if node.cat == LatCat.CMD and node.token == 'mid' and not has_sibling(node, lambda x: r'\{' in x.token):
            add_error('MID_IN_SET_COMPREHENSION', node)

        if (node.cat == LatCat.CMD and node.token in ['sum', 'prod', 'frac', 'binom']
                and node.prev_node and node.prev_node.is_in_math and node.prev_node.cat != LatCat.CMD and not node.prev_node.is_in_arg and re_ends_with_digit_or_letter.match(node.prev_node.token)):
            add_error('CDOT_FOR_READABILITY', node)
        if node.cat == LatCat.CMD and node.token == 'centering':
            if not (node.parent and node.parent.token in ['figure', 'table']):
                add_error('CENTERING', node)
        if node.cat == LatCat.CMD and node.token == 'not' and node.next_node:
            if (node.next_node.token.strip()+' ')[0] == '=' or node.next_node.token in ['in']:
                add_error('INCORPORATE_NOT', node)
        if node.cat == LatCat.CMD and node.token == 'over':
            add_error('OVER_VS_FRAC', node)
        if node.cat == LatCat.CMD and node.token == 'choose':
            add_error('CHOOSE_VS_BINOM', node)
        if (node.cat == LatCat.CMD and node.token in ['in', 'notin', 'ni', 'subset', 'subseteq']
                and node.next_node and node.next_node.token and (node.next_node.token.strip()+' ')[0] in 'NRZQC'):
            add_error('SETS_IN_BBFONT', node)

        if node.cat == LatCat.CMD and node.token == 'label' and node.args and len(node.args) == 1 and re_trivial_label.match(node.args[0].tex):
            add_error('TRIVIAL_LABEL', node)

        if node.cat == LatCat.CMD and node.token == 'ref' and node.prev_node and node.prev_node.token and node.prev_node.token.endswith('('):
            add_error('EQREF_INSTEAD_OF_REF', node)
        if node.cat == LatCat.CMD and node.token in ['ref', 'eqref'] and node.prev_node and node.prev_node.token and re_ends_with_space.search(node.prev_node.token):
            add_error('NONBREAKABLE_SPACE_BEFORE_REF', node)

        if node.is_in_display_math and node.token == 'limits':
            add_error('LIMITS_UNNECESSARY_IN_DISPLAY_MODE', node)

        if node.is_in_math and node.token in ['mbox', 'hbox']:
            add_error('REPLACE_MBOX_WITH_TEXT', node)

        if node.is_in_math and node.cat == LatCat.CMD and node.token in ['textbf', 'textit']:
            add_error('TEXT_COMMANDS_IN_MATH_MODE')

        if node.is_in_math and node.cat == LatCat.STR:
            if '|' in node.token and (r'\{' in node.token or has_sibling(node, lambda x: r'\{' in (x.token or ''))) and not has_sibling(node, lambda x: x.token == 'mid'):
                add_error('MID_IN_SET_COMPREHENSION', node)
            if '(' in node.token and node.next_node and node.next_node.token in ['sum', 'prod', 'frac', 'binom']:
                add_error('LEFT_RIGHT_RECOMMENDED', node)
            if re_mod_cmd.search(node.token):
                add_error('MOD_NOT_A_COMMAND', node)
            if '...' in node.token:
                add_error('ELLIPSIS_LDOTS', node)
            if re_possibly_word.search(node.token):
                add_error('TEXT_IN_MATH_MODE', node)
            if '>=' in node.token or '<=' in node.token:
                add_error('LE_AS_SINGLE_COMMAND', node)
            if re_math_no_backslash.search(node.token):
                add_error('BACKSLASH_NEEDED', node)
            if '√' in node.token:
                add_error('UNICODE_SQRT', node)
            if re_cyrillic_tricky_letter.search(node.token):
                add_error('CYRILLIC_LETTER_C_MISUSED', node)
            if '"' in node.token:
                add_error('QUOTES_IN_MATH', node)
            if node.token.strip().endswith('-') and node.next_node and not node.next_node.is_in_math:
                add_error('DASH_IN_MATH_MODE', node)
            if node.token.strip() == '' and node.prev_sibling and node.prev_sibling.cat == LatCat.ENV and node.prev_sibling.token == 'math' and node.next_sibling and node.next_sibling.cat == LatCat.ENV and node.next_sibling.token == 'math':
                add_error('UNNECESSARY_FORMULA_BREAK', node)

        if not node.is_in_math and node.cat == LatCat.STR and not is_pure_argument(node):
            if re_nonsymbolic_reference.search(node.token):
                add_error('SYMBOLIC_LINKS')
            if re_starts_with_uppercase.match(node.token) and node.prev_node.is_in_math and not (node.prev_node.cat == LatCat.STR and node.prev_node.token.strip().endswith('.')):
                add_error('PERIOD_BEFORE_NEXT_SENTENCE')
            if '~ ' in node.token or ' ~' in node.token:
                add_error('TILDE_INEFFECTIVE_AS_NBSP', node)
            if re_capitalization_after_comma.search(node.token) or (re_starts_with_uppercase.match(node.token) and node.prev_node and node.prev_node.cat == LatCat.STR and node.prev_node.token.strip().endswith(',')):
                add_error('CAPITALIZATION_AFTER_PUNCTUATION_MARK', node)
            if re_capitalization_after_period.search(node.token) or (re_starts_with_lowercase.match(node.token) and node.prev_node and node.prev_node.cat == LatCat.STR and node.prev_node.token.strip().endswith('.')):
                add_error('CAPITALIZATION_AFTER_PERIOD', node)
            if '...' in node.token:
                # maybe make two different errors for math and text mode
                add_error('ELLIPSIS_LDOTS', node)
            if (node.token.strip() + ' ')[0] in ',.:?!;' and is_display_math(node.prev_sibling):
                add_error('PUNCTUATION_AFTER_DISPLAY_MATH', node.prev_sibling)
            if re_space_before_parenthesis.search(node.token):
                add_error('SPACE_BEFORE_PARENTHESIS', node)
            if re_space_after_parenthesis.search(node.token):
                add_error('SPACE_AFTER_PARENTHESIS', node)
            if re_space_after_punctuation.search(node.token):
                add_error('SPACE_AFTER_PUNCTUATION_MARK', node)
            if re_space_before_punctuation.search(node.token):
                add_error('SPACE_BEFORE_PUNCTUATION_MARK', node)
            if re_latin_letter_outside_math_ru.search(node.token):
                add_error('LATIN_LETTER_OUTSIDE_MATH_RU', node)
            if re_latin_letter_outside_math_en.search(node.token):
                add_error('LATIN_LETTER_OUTSIDE_MATH_EN', node)
            if re_math_command.search(node.token):
                add_error('MATH_SEMANTICS_OUTSIDE_MATH', node)
            if re_latin_c_in_rus_text.search(node.token):
                add_error('LATIN_LETTER_C_MISUSED', node)
            if '"' in node.token and node.token != r'\"':
                add_error('WRONG_QUOTES', node)
            if "''" in node.token and "``" not in node.token or "``" in node.token and "''" not in node.token:
                add_error('WRONG_SAME_QUOTES', node)
            if re_dash_as_hyphen.search(node.token):
                add_error('DASH_HYPHEN', node)
            if re_td.search(node.token):
                add_error('ABBREVIATIONS_WITH_SPACE', node)
            if re_dash_no_spaces.search(node.token):
                add_error('DASH_SURROUND_WITH_SPACES', node)
            if re_ru_ordinal.search(node.token):
                add_error('RU_ORDINAL_ABBREVIATION', node)
            if '\\\\' in node.token:
                add_error('SUGGESTED_NEW_PARAGRAPH', node)
            if re_small_numeral.search(node.token):
                add_error('NUMERALS_AS_WORDS', node)
        if (node.token and node.token.strip().endswith(r'\\') or node.token == r'par') and node.next_node and is_display_math(node.next_node):
            add_error('PARAGRAPH_BREAK_BEFORE_DISPLAY_FORMULA', node)

        if node.cat == LatCat.ENV and node.token == 'math' and node.next_sibling and node.next_sibling.cat == LatCat.ENV and node.next_sibling.token == 'math':
            add_error('UNNECESSARY_FORMULA_BREAK', node)

        if node.cat == LatCat.CMD and node.token in ['it', 'bf', 'sf', 'rm']:
            add_error('LOW_LEVEL_FONT_COMMANDS', node)

        if node.cat == LatCat.CMD and node.token == 'textit':
            add_error('ITALIC_INSTEAD_OF_EMPH', node)

        if node.cat == LatCat.CMD and node.token == 'cite' and (
                node.prev_node and (node.prev_node.is_in_math or is_math(node.prev_node))
                or
                node.next_sibling and is_math(node.next_sibling)
                ):
            add_error('FORMULA_NEIGHBOURING_REFERENCE', node)
        if node.is_in_math and node.token == 'includegraphics':
            add_error('GRAPHICS_IN_MATH_MODE', node)

    for node in all_nodes:
        if (node.cat == LatCat.ENV) and is_display_math(node.token) and node.next_sibling and (
                node.next_sibling.cat == LatCat.ENV) and is_display_math(node.next_sibling.token):
            add_error('CONSECUTIVE_DISPLAY_FORMULAE', node.next_sibling.pos or node.pos)

    used_labels = set()
    for node in all_nodes:
        if node.cat == LatCat.CMD and node.token in ['ref', 'eqref'] and node.args and len(node.args) == 1:
            used_labels.add(node.args[0].tex)
    for node in all_nodes:
        if node.cat == LatCat.CMD and node.token in ['label'] and node.args and len(node.args) == 1 and node.args[0].tex not in used_labels:
            add_error('NUMBERED_MATH_NEEDS_REFERENCING', node)

    return errors


def html_to_console(html_text, rich_text=True, width=80):
    formats = {
        'red': '\033[91m',
        'green': '\033[92m',
        'blue': '\033[94m',
        'bold': '\033[1m',
        'italics': '\033[3m',
        'underline': '\033[4m',
        'end': '\033[0m'
    }
    if not rich_text:
        formats = {k: '' for k in formats}

    html_entities = {
        '&nbsp;': ' ',
        '&thinsp;': '\u2009',
        '&amp;': '&',
        '&mdash;': '—',
        '&ndash;': '–',
        '&hellip;': '…',
        '&laquo;': '«',
        '&raquo;': '»',
        '&lt;': '<',
        '&gt;': '>'
    }

    text = html_text
    for entity, char in html_entities.items():
        text = text.replace(entity, char)

    wrapper = textwrap.TextWrapper(width=width)
    text = wrapper.fill(text)

    text = re.sub(r'<em>(.*?)</em>', f"{formats['italics']}\\1{formats['end']}", text)
    text = re.sub(r'<strong>(.*?)</strong>', f"{formats['bold']}\\1{formats['end']}", text)
    text = re.sub(r'<code>(.*?)</code>', f"{formats['green']}\\1{formats['end']}", text)
    text = re.sub(r'<a\s+href="([^"]*?)"[^>]*>(.*?)</a>', f"[{formats['underline']}\\2{formats['end']}](\\1)", text)

    return text


def main():
    parser = argparse.ArgumentParser(description='Perform checks on a LaTeX file.')
    parser.add_argument('filename', help='Path to the LaTeX file')
    parser.add_argument('-l', '--language', choices=['EN', 'RU'], default='EN', help='Language for error descriptions (EN or RU)')
    parser.add_argument('-r', '--rich', choices=['Y', 'N'], help='Enable rich text in console output', default='Y')
    args = parser.parse_args()

    filename = args.filename
    if not os.path.exists(filename):
        print(f"'File {filename} does not exist.")
        return
    with open(filename, 'r') as infile:
        source = infile.read()

    language = SupportedLanguages.EN if args.language.lower() == 'en' else SupportedLanguages.RU
    rich = args.rich.lower() == 'y'
    errors = perform_checks(source, language=language, debug_mode=False)

    for key, value in errors.items():
        print(f'{key}:\n    {"\n    ".join(repr(source[max(0, t-10):t+20]) for t in value)}')
        if explanation := error_descriptions[language][key]:
            print(f'\nExplanation:  {html_to_console(explanation['msg'], rich_text=rich)}\n')


if __name__ == '__main__':
    main()



from enum import Enum


class SupportedLanguages(Enum):
    RU = 'RU'
    EN = 'EN'


error_descriptions = {}

error_descriptions[SupportedLanguages.RU] = {
    "PARSE_ERROR": {
        "msg": "Ошибка при разборе LaTeX-кода. Возможно, где-то не закрыта формула знаком доллара, либо наоборот забыт доллар перед формулой.",
        "severity": 0
    },
    "DOUBLE_DOLLARS": {
        "msg": "Двойных долларов в тексте быть не должно. Выносные формулы нужно оформлять с помощью <code>\\[</code>…<code>\\]</code>. Объяснение <a href=\"https://tex.stackexchange.com/questions/503/why-is-preferable-to\" target=\"_blank\">по ссылке</a>.",
        "severity": 0
    },
    "CONSECUTIVE_DISPLAY_FORMULAE": {
        "msg": "Обнаружены две идущие подряд выключные формулы. Такого быть не должно: используйте окружение <code>aligned</code> или его аналоги, чтобы грамотно оформить не умещающиеся на одной строке выкладки. Подробнее, например, по <a href=\"https://www.overleaf.com/learn/latex/Aligning_equations_with_amsmath\" target=\"_blank\">ссылке</a>.",
        "severity": 0
    },
    'LINEBREAK_AFTER_DISPLAY_FORMULAE': {
        "msg": "Обнаружен разрыв строки непосредственно рядом с выключной формулой. После выключной формулы переход на новую строку выполняется автоматически. Если Вы хотели увеличить отступ по вертикали до следующего за формулой текста, используйте другие средства.",
        "severity": 0
    },
    'LINEBREAK_BEFORE_DISPLAY_FORMULAE': {
        "msg": "Обнаружен разрыв строки непосредственно перед выключной формулой. Перед выключной формулой переход на новую строку выполняется автоматически. Если Вы хотели увеличить отступ по вертикали между текстом и формулой.",
        "severity": 0
    },
    "EQNARRAY_USED": {
        "msg": "Не используйте окружение <code>eqnarray</code> (подробности по <a href=\"https://tex.stackexchange.com/a/197\" target=\"_blank\">ссылке</a>). Вместо этого пользуйтесь, например, <code>align</code>.",
        "severity": 0
    },
    "ABBREVIATIONS_WITH_SPACE": {
        "msg": "Сокращения типа <em>т.&thinsp;е.</em>, <em>т.&thinsp;к.</em>, <a href=\"https://ru.wikipedia.org/wiki/Q.E.D.\" target=\"_blank\"><em>ч.&thinsp;т.&thinsp;д.</em></a> и подобные <a href=\"https://new.gramota.ru/spravka/buro/search-answer?s=296030\">следует оформлять с пробелом</a> (см. также <a href=\"https://popravilam.com/blog/105-probel-v-sokrashcheniyah.html\" target=\"_blank\">тут</a>), но есть особенность: использовать нужно <em>неразрывный</em> пробел <code>~</code> или, ещё лучше, <a href=\"https://ru.wikipedia.org/wiki/%D0%A3%D0%B7%D0%BA%D0%B8%D0%B9_%D0%BF%D1%80%D0%BE%D0%B1%D0%B5%D0%BB\" target=\"_blank\"><em>тонкую шпацию</em></a> <code>\\,</code> — это неразрывный пробел несколько меньшей ширины, нежели обычный. Например: <code>ч.\\,т.\\,д.</code>. В противном случае может случиться казус при переносе текста, когда часть сокращения останется на строке, а часть перейдёт на следующую. Тонкая шпация также смотрится лучше в этом случае, чем обычный пробел. Также её используют при наборе <a href=\"https://new.gramota.ru/spravka/letters/78-init\" target=\"_blank\">инициалов</a>, например: <code>М.\\,В.~Ломоносов</code> или <code>М.\\,В.\\,Ломоносов</code>.",
        "severity": 0
    },
    "DASH_HYPHEN": {
        "msg": "Возможно, перепутано тире с дефисом. Полноценное длинное тире ставится с помощью <code>---</code>, укороченное с помощью <code>--</code>. Подробнее о тире, дефисах и подобном см. <a href=\"https://webstyle.sfu-kras.ru/tire-defis\" target=\"_blank\">здесь</a> и <a href=\"https://habrahabr.ru/post/20588/\" target=\"_blank\">здесь</a>. Ну и, конечно, никогда не поздно почитать <a href=\"https://www.artlebedev.ru/kovodstvo/sections/97/\" target=\"_blank\">А. Лебедева</a>.",
        "severity": 0
    },
    "DASH_IN_MATH_MODE": {
        "msg": "Похоже, Вы хотели поставить дефис. Но когда знак дефиса попадает в математический режим, он становится минусом. Вывод: делать дефис не частью формулы, а частью следующего за ней текста.",
        "severity": 0
    },
    "DASH_SURROUND_WITH_SPACES": {
        "msg": "Тире с двух сторон следует окружать пробелами. Особенный шик — один или оба из пробелов сделать неразрывными, чтобы тире не «повисало на краю пропасти» при переносе строки. Подробнее о тире, дефисах и подобном см. <a href=\"https://webstyle.sfu-kras.ru/tire-defis\" target=\"_blank\">здесь</a> и <a href=\"https://habrahabr.ru/post/20588/\" target=\"_blank\">здесь</a>. Ну и, конечно, никогда не поздно почитать <a href=\"https://www.artlebedev.ru/kovodstvo/sections/97/\" target=\"_blank\">А. Лебедева</a>.",
        "severity": 0
    },
    "RU_ORDINAL_ABBREVIATION": {
        "msg": "Нарушены <a href=\"https://new.gramota.ru/spravka/letters/22-spravka/letters/87-rubric-99\" target=\"_blank\">правила сокращения порядковых числительных</a>.",
        "severity": 0
    },
    "PARAGRAPH_BREAK_BEFORE_DISPLAY_FORMULA": {
        "msg": "Пустая строка заставляет \\(\\LaTeX\\) начинать новый параграф, даже если эта пустая строка стоит прямо перед выключной формулой. Чаще всего это не нужно, так как параграфы по смыслу не следует начинать с формул, а выключную формулу \\(\\LaTeX\\) в любом случае окаймляет достаточным количеством пустого пространства.",
        "severity": 0
    },
    "UNNECESSARY_FORMULA_BREAK": {
        "msg": "Возможно, некоторые формулы следовало объединить. Например, вместо <code>$x$, $y$ $\\in$ $A$</code> пишите <code>$x,\\,y\\in A$</code>; вместо <code>$a$ = $b+c$</code> пишите <code>$a=b+c$</code> и т.д.",
        "severity": 0
    },
    "CENTERING": {
        "msg": "Центрирование явно используется только при включении рисунков и таблиц. Иногда для заголовков. Для обычных формул центрирование не используется, вместо этого следует делать выключные формулы.",
        "severity": 0
    },
    "LOW_LEVEL_FONT_COMMANDS": {
        "msg": "Вместо низкоуровневых команд <code>{\\it …}</code>, <code>{\\bf …}</code> используйте команды <code>\\textit{…}</code> <code>\\textbf{…}</code> (подробности см. по <a href=\"https://tex.stackexchange.com/questions/41681/correct-way-to-bold-italicize-text\" target=\"_blank\">ссылке</a>.) Кроме того, выделять текст лучше не курсивом, а более гибкой командой <code>\\emph{…}</code>, поскольку она корректно сработает даже внутри курсивного блока.",
        "severity": 0
    },
    "WRONG_QUOTES": {
        "msg": "Для закавычивания слов «клавиатурные» кавычки <code>\"…\"</code> в \\(\\LaTeX\\) не используются. Если нужно закавычить слово, делайте это одним из способов <code>``…''</code> (два апострофа на букве ё вначале и два на букве э в конце) или <code><<…>></code> (два знака меньше и два знака больше).",
        "severity": 0
    },
    "WRONG_SAME_QUOTES": {
        "msg": "Открывающие и закрывающие кавычки ставятся по-разному: непривильно делать <code>''…''</code> или <code>``…``</code>, следует делать <code>``…''</code> для закавычивания в англоязычном тексте. В русской типографике применяются французские <code><<кавычки-ёлочки>></code>, для вложенных кавычек — немецкие <code>,,кавычки-лапки``</code>.",
        "severity": 0
    },
    "QUOTES_IN_MATH": {
        "msg": "Обнаружен символ кавычки в математическом режиме. Если Вы хотели поставить двойной штрих над математическим символом, наберите два штриха подряд: <code>y''</code>, либо используйте команду <code>\\prime\\prime</code>.",
        "severity": 0
    },
    "LATIN_LETTER_OUTSIDE_MATH_RU": {
        "msg": "Даже одна буква, если у неё математический смысл (если это имя математического объекта) — должна быть набрана в математическом режиме.",
        "severity": 0
    },
    "MATH_SEMANTICS_OUTSIDE_MATH": {
        "msg": "Команды, печатающие символы, имеющие математическую природу, настоятельно рекомендуется использовать в математическом режиме, даже если они работают и без оного.",
        "severity": 0
    },
    "LATIN_LETTER_C_MISUSED": {
        "msg": "Возможно, использована случайно латинская буква с (це) вместо русского предлога <strong>с</strong> посреди текста на русском языке.",
        "severity": 0
    },
    "CYRILLIC_LETTER_C_MISUSED": {
        "msg": "Возможно, использована случайно одна из кириллических букв <strong>а</strong>, <strong>е</strong>, <strong>с</strong>, <strong>х</strong> вместо соответствующей латинской буквы внутри формулы.",
        "severity": 0
    },
    "MULTIPLICATION_SIGN": {
        "msg": "Знак <code>*</code> используется для обозначения умножения в программировании, но не в математике. Пользуйтесь командой <code>\\cdot</code> или <code>\\times</code> (последней только в особых случаях).",
        "severity": 0
    },
    "SPACE_BEFORE_PUNCTUATION_MARK": {
        "msg": "Пробелы перед двоеточием, запятой, точкой, восклицательным и вопросительными знаками, точкой с запятой не ставятся.",
        "severity": 0
    },
    "SPACE_BEFORE_PARENTHESIS": {
        "msg": "В тексте перед открывающей скобкой ставится пробел.",
        "severity": 0
    },
    "SPACE_AFTER_PUNCTUATION_MARK": {
        "msg": "После двоеточия, запятой, точки, точки с запятой, восклицательного и вопросительного знаков нужно ставить пробел.",
        "severity": 0
    },
    "SPACE_AFTER_PARENTHESIS": {
        "msg": "После открывающей скобки не следует ставить пробел.",
        "severity": 0
    },
    "CAPITALIZATION_AFTER_PUNCTUATION_MARK": {
        "msg": "После двоеточия, точки с запятой, запятой, — не должно должно быть заглавной буквы, поскольку предложение продолжается.",
        "severity": 0
    },
    "CAPITALIZATION_AFTER_PERIOD": {
        "msg": "Предложение следует начинать с заглавной буквы.",
        "severity": 0
    },
    "PERIOD_BEFORE_NEXT_SENTENCE": {
        "msg": "В конце предложения должна ставиться точка, даже если предложение заканчивается формулой.",
        "severity": 0
    },
    "LEFT_RIGHT_RECOMMENDED": {
        "msg": "Когда при наборе формул возникает необходимость поставить скобки (круглые/фигурные/квадратные) или знак модуля вокруг высокой подформулы (дроби, биномиального коэффициента, суммы с пределами), рекомендуется добавлять команды <code>\\left</code> и <code>\\right</code>. Особенно это актуально для выключных формул. Например, сравните, как нелепо выглядит в PDF скомпилированная формула <code>\\[(\\frac{a}{b})^2\\]</code> и как естественен её «правильный» вариант <code>\\[\\left(\\frac{a}{b}\\right)^2\\]</code>. Тем не менее, переусердствовать здесь тоже не стоит: добавляйте <code>\\left… … \\right…</code> только тогда, когда видите явное несоответствие между высотой скобок и высотой того, что они окружают. Подробности по <a href=\"https://tex.stackexchange.com/a/58641\" target=\"_blank\">ссылке</a>. Можно также воспользоваться командой <a href=\"https://tex.stackexchange.com/a/1765\" target=\"_blank\"><code>\\DeclarePairedDelimiter</code></a> из библиотеки mathtools.",
        "severity": 10
    },
    "SYMBOLIC_LINKS": {
        "msg": "Пользуйтесь символическими ссылками, например: <code>Заметим, что \\begin{equation}\\label{eqSquares} a^2-b^2=(a-b)(a+b). \\end{equation} Из равенства~\\eqref{eqSquares}, следует, что…</code>.",
        "severity": 0
    },
    "EQREF_INSTEAD_OF_REF": {
        "msg": "Вместо <code>(\\ref{…})</code> следует писать <code>\\eqref{…}</code>",
        "severity": 0
    },
    "NONBREAKABLE_SPACE_BEFORE_REF": {
        "msg": "Перед командами <code>\\ref</code>, <code>\\eqref</code> рекомендуется ставить тильду, чтобы номер ссылки был приклеен неразрывным пробелом, например: <code>…Согласно~\\eqref{eqMaxwell}, имеем…</code>.",
        "severity": 0
    },
    "ELLIPSIS_LDOTS": {
        "msg": "В математическом режиме многоточие ставится не <code>...</code>, а командой <code>\\ldots</code>, а в текстовом режиме — командой <code>\\textellipsis</code>. При наличии пакета amsmath, когда многоточие нужно вставить между запятыми в перечислении объектов, используйте команду <code>\\dotsc</code>.",
        "severity": 0
    },
    "TRIVIAL_LABEL": {
        "msg": "Символические ссылки нужно делать, но нет большого смысла делать тривиальные описания типа <code>\\label{eq1}</code>. Так же, как в программировании называть переменную <code>var1</code> чаще всего пагубно. Куда лучше придумать осмысленное название, например <code>eqCauchy</code>, <code>eqBinomialSymmetry</code>, <code>eqMain</code> (если это действительно самая важная формула в доказательстве) и т.д.",
        "severity": 5
    },
    "REPLACE_MBOX_WITH_TEXT": {
        "msg": "Для вставки текста внутрь формулы вместо команд <code>\\mbox</code> и <code>\\hbox</code> пользуйтесь командой <code>\\text</code>. Объяснение см. по <a href=\"https://tex.stackexchange.com/questions/70632/difference-between-various-methods-for-producing-text-in-math-mode\" target=\"_blank\">ссылке</a>.",
        "severity": 0
    },
    "TEXT_IN_MATH_MODE": {
        "msg": "Для вставки текста внутрь формулы (даже если слова на латинице, но не являются именами математических объектов) вместо команд пользуйтесь командой <code>\\text</code>, например: <code>Рассмотрим множество $A_{\\text{хорошие}}$</code>.",
        "severity": 0
    },
    "INCORPORATE_NOT": {
        "msg": "Хотя префикс <code>\\not</code> позволяет из многих значков получить значок-отрицание, часто короче (и <a href=\"https://tex.stackexchange.com/a/141011\" target=\"_blank\">рекомендуется</a>!) писать одной командой. Например, вместо <code>\\not=</code> можно написать <code>\\ne</code>, вместо <code>\\not\\in</code> написать <code>\\notin</code> и т.д.",
        "severity": 0
    },
    "OVER_VS_FRAC": {
        "msg": "Команда <code>\\over</code> является низкоуровневой командой TeX и <a href=\"https://tex.stackexchange.com/a/73825\" target=\"_blank\">не рекомендуется</a> к использованию в \\(\\LaTeX\\). Вместо <code>A \\over B</code> пишите <code>\\frac{A}{B}</code>.",
        "severity": 0
    },
    "CHOOSE_VS_BINOM": {
        "msg": "Команда <code>\\choose</code> является низкоуровневой командой TeX и <a href=\"https://tex.stackexchange.com/a/127711\" target=\"_blank\">не рекомендуется</a> к использованию в \\(\\LaTeX\\). Вместо <code>A \\choose B</code> пишите <code>\\binom{A}{B}</code>.",
        "severity": 0
    },
    "SETS_IN_BBFONT": {
        "msg": "Стандартные числовые множества (натуральные числа и пр.) нужно набирать специальным шрифтом: вместо <code>N</code> пишите <code>\\mathbb{N}</code>.",
        "severity": 0
    },
    "MOD_NOT_A_COMMAND": {
        "msg": "Используйте команду <code>\\bmod</code> или <code>\\pmod</code>, чтобы mod был набран в формуле прямым шрифтом (как и подобает операции, а не произведению трёх переменных m, o и d).",
        "severity": 0
    },
    "TILDE_INEFFECTIVE_AS_NBSP": {
        "msg": "Значок <code>~</code> означает неразрывный пробел. Окружать его пробелами бессмысленно: неразрывность пропадает. Например, вместо <code>По формуле ~\\eqref{…</code> следует писать <code>По формуле~\\eqref{…</code>",
        "severity": 0
    },
    "INDENTATION_WITH_SPACES": {
        "msg": "Избегайте использования нескольких пробельных значков подряд, так же, как и в WYSIWYG-редакторах не следует делать отступы множественными пробелами. Выберите <em>одну</em> подходящую команду \\(\\LaTeX\\) для отступа в конкретной ситуации, подробности по <a href=\"https://tex.stackexchange.com/a/74354\" target=\"_blank\">ссылке</a>.",
        "severity": 0
    },
    "LE_AS_SINGLE_COMMAND": {
        "msg": "Чтобы набрать символ \\(\\le\\) или \\(\\ge\\), используйте команды <code>\\le</code> и <code>\\ge</code> соответственно. Стрелка \\(\\Leftarrow\\) набирается командой <code>\\Leftarrow</code>. Программистские сочетания <code><=</code> и <code>>=</code> ни в одной из ситуаций не годятся.",
        "severity": 0
    },
    "PUNCTUATION_AFTER_DISPLAY_MATH": {
        "msg": "Если поставить знак препинания после выключной формулы, он будет отображён на другой строке. Поэтому при необходимости поставить, например, запятую сразу после выключной формулы, запятую следует сделать частью этой формулы.",
        "severity": 0
    },
    "BACKSLASH_NEEDED": {
        "msg": "Слова \\(\\min\\), \\(\\max\\) и подобные в формулах являются именами математических операторов и должны набираться прямым шрифтом. В \\(\\LaTeX\\) есть команды <code>\\min</code>, <code>\\max</code>, <code>\\lim</code>, <code>\\deg</code>, и другие, которые делают эту работу за Вас. Список доступных стандартных команд см. по <a href=\"https://www.overleaf.com/learn/latex/Operators#Reference_guide\" target=\"_blank\">ссылке</a>. Если такой команды ещё нет, используйте конструкцию типа <code>\\operatorname{min}</code> или, ещё лучше, <a href=\"https://tex.stackexchange.com/a/67529\" target=\"_blank\">создайте свой оператор</a> командой <code>\\DeclareMathOperator</code> в преамбуле документа.",
        "severity": 0
    },
    "CDOT_FOR_READABILITY": {
        "msg": "В произведениях чисел, обозначаемых отдельной буквой, и дробей или биномиальных коэффициентов, полезно для улучшения читабельности текста явно указывать произведение командой <code>\\cdot</code>.",
        "severity": 5
    },
    "GRAPHICS_IN_MATH_MODE": {
        "msg": "Команда <code>\\includegraphics</code> не должна использоваться в математическом режиме без <em>крайней</em> необходимости. Чтобы отцентрировать рисунок, вместо помещения рисунка в выключную формулу используйте окружения <code>center</code> и <code>figure</code>.",
        "severity": 0
    },
    "UNNECESSARY_MATH_MODE": {
        "msg": "Если внутри формулы (в окружении долларов) стоит единственный символ и он не является буквой или цифрой — это тревожный знак. Скорее всего, либо в математический режим переходить было не нужно, либо без нужды на части была разорвана формула.",
        "severity": 0
    },
    "NO_SPACE_AFTER_COMMAND_BEFORE_CYRILLIC": {
        "msg": "Хотя \\(\\LaTeX\\) это и не считает ошибкой, не следует писать слитно команды ТеХа и кириллические слова.",
        "severity": 0
    },
    "TEXT_COMMANDS_IN_MATH_MODE": {
        "msg": "Текстовые команды <code>\\textbf</code>, <code>\\textit</code> не следует использовать в математическом режиме. Для набора жирным шрифтом математических символов есть команда <code>\\mathbf</code>.",
        "severity": 0
    },
    "LIMITS_UNNECESSARY_IN_DISPLAY_MODE": {
        "msg": "Команда <code>\\limits</code> в выключных формулах обычно лишняя: пределы и без неё выставляются верно.",
        "severity": 0
    },
    "FORMULA_NEIGHBOURING_REFERENCE": {
        "msg": "Плохо читается, когда формула непосредственно соседствует со ссылкой, без знаков препинания. Всегда можно вставить слово либо поменять порядок слов в предложении, чтобы этого избежать.",
        "severity": 0
    },
    "UNICODE_SQRT": {
        "msg": "Для квадратного корня следует использовать команду <code>\\sqrt</code> вместо символа Unicode.",
        "severity": 0
    },
    "NUMBERED_MATH_NEEDS_REFERENCING": {
        "msg": "Если у формулы есть видимый номер, то этот номер должен быть использован где-то в тексте для ссылки на формулу. А если на формулу необязательно ссылаться, то и нумеровать её не следует.",
        "severity": 0
    },
    # todo
    "NO_CONCLUSION": {
        "msg": "Обычно решение задачи или доказательство теоремы не заканчивают рисунком или формулой. Полезно добавить в конце хотя бы какое-то заключение, например: <code>Теорема доказана.</code>, <code>В итоге мы получили ответ: искомое количество равно $42$.</code> и т.д.",
        "severity": 5
    },
    "SUGGESTED_NEW_PARAGRAPH": {
        "msg": "Чтобы начать с новой строки формулу, она делается выключной. Чтобы начать новый параграф текста, используется команда <code>\\par</code> в текстовом режиме. Использовать же несемантичный перенос <code>\\\\</code>, помимо окружений типа <code>tabular</code>, следует только в крайних случаях.",
        "severity": 0
    },
    "NUMERALS_AS_WORDS": {
        "msg": "Со стилистической точки зрения, числительные, не превосходящие пяти и окружённые текстом, часто лучше писать не цифрами, а словами. Например, <em>Рассмотрим 2 случая</em> смотрится хуже, чем <em>Рассмотрим два случая</em>.",
        "severity": 10
    },
    "MID_IN_SET_COMPREHENSION": {
        "msg": "При описании множеств вертикальная черта ставится командой <code>\\mid</code>. Например, <code>\\{x^2\\mid x\\in\\mathbb{Z}\\}</code>. И, наоборот, <code>\\mid</code> НЕ используется в остальных контекстах, например, при обозначении модуля числа. Для обозначения последнего пишите <code>|x|</code> или <code>\\lvert x \\rvert</code>.",
        "severity": 0
    },
    # todo
    "FLOOR_FUNCTION_NOTATION": {
        "msg": "В современной литературе принято целую часть числа \\(x\\) обозначать <code>\\lfloor x\\rfloor</code>, а не <code>[x]</code>.",
        "severity": 0
    },
    "ITALIC_INSTEAD_OF_EMPH": {
        "msg": "Выделять текст лучше не курсивом, а более гибкой командой <code>\\emph{…}</code>, поскольку она корректно сработает даже внутри курсивного блока.",
        "severity": 0
    },
    # todo
    "PARAGRAPH_STARTS_WITH_FORMULA": {
        "msg": "Параграф не следует начинать с формулы. Если Вы хотите, чтобы формула была на отдельной строке, а не в тексте, сделайте её выносной: <code>\\[…\\]</code>.",
        "severity": 0
    },
    # todo
    "SENTENCE_STARTS_WITH_FORMULA": {
        "msg": "Обычно предложение не следует начинать с формулы. Почти всегда можно добавить в начало слово. Например, вместо предложения <code>$G$ связен.</code> напишите <code>Граф $G$ связен.</code>",
        "severity": 0
    },
    # todo
    "SENTENCE_STARTS_WITH_NUMBER": {
        "msg": "Обычно предложение не следует начинать с цифровой записи числа. Например, вместо предложения <code>5 человек можно выстроить в очередь $5!$ способами.</code> лучше написать либо <code>Пятерых человек…</code> либо <code>Заметим, что 5 человек…</code>",
        "severity": 5
    },
    # todo
    "RUSSIAN_TYPOGRAPHY_PECULIARITIES": {
        "msg": "Строго говоря, это не является ошибкой, но в отечественной типографской традиции принято пустое множество обозначать значком <code>\\varnothing</code>, а не <code>\\emptyset</code> (последний более приплюснутый). Аналогично, вместо <code>\\epsilon</code> лучше писать <code>\\varepsilon</code>, а вместо <code>\\phi</code> писать <code>\\varphi</code>.",
        "severity": 0
    },
    # todo
    "BETTER_TO_USE_WORDS_THEN_MATH": {
        "msg": "Конструкции наподобие <code>число элементов $=m^2$</code> недопустимы в письменном тексте, за исключением конспектов. Знаки \\(=, \\gt, \\geqslant\\) и др. нужно в этих случаях писать словами: <code>…не превосходит $m^2$</code>, <code>…равняется $m^2$</code> и т.д.",
        "severity": 0
    },
    # todo
    "MATH_ENVIRONMENT_VERBOSITY_WARNING": {
        "msg": "Окружение <code>math</code> используется довольно редко. Лучше всего вместо него использовать более короткую (и абсолютно такую же по получаемому результату) конструкцию <code>\\(…\\)</code> или <code>$…$</code>. Часто также ошибочно полагают, что с помощью math оформляют <em>выключные</em> формулы — но это не так: внутри окружения math действует обычный inline-режим. Для оформления выключных формул подойдёт либо конструкция <code>\\[…\\]</code>, либо одно из окружений equation, array и др.",
        "severity": 0
    },
    # todo
    "USE_DIVIDES_INSTEAD_OF_VDOTS": {
        "msg": "Команду <code>\\vdots</code> следует использовать только в матрицах или похожих окружениях для обозначения именно многоточия по вертикали. Когда речь идёт о делимости, вместо трёх точек (в качестве слов «делится на») используйте вертикальную черту (в качестве слов «является делителем»). Когда же рядом ещё вертикальные черта, например, в set builder notation, разумно вообще писать словами: <code>\\( A=\\{x\\in\\mathbb{N} \\mid x\\text{ кратен } 5\\} \\)</code>",
        "severity": 0
    },
    # todo
    "MAKE_LONG_FORMULA_DISPLAY": {
        "msg": "Подозрительно длинная формула набрана не в выключном режиме. Формулы, которые при компиляции не влезают целиком на одну строку (т.е. вся строка занята формулой и всё равно возникает перенос), нужно делать выключными, аккуратно их разбивая построчно с помощью окружений AMS: см. перечень подходящих окружений в <a href=\"https://www.overleaf.com/learn/latex/Aligning_equations_with_amsmath\" target=\"_blank\">документе по ссылке</a>.",
        "severity": 0
    },
    # todo
    "EN_ORDINAL_ABBREVIATION": {
        "msg": "",
        "severity": 0
    },
    # todo
    "LATE_DEFINITION": {
        "msg": "Вместо того, чтобы писать <q><code>$x=a+b$</code>, где <code>$a=…</code></q> сначала лучше ввести все буквы и лишь затем записать выражение, эти буквы содержащие. См. <a href=\"https://1drv.ms/w/s!AiAwrmxQ9QLrjOtTQqAXWdl3ryK5Jg?e=vEMU6W\" target=\"_blank\">статью П. Халмоша</a>, раздел «Правильно используйте слова».",
        "severity": 0
    },
    # todo
    "INVISIBLE_BRACES": {
        "msg": "Чтобы вывести на экран фигурные скобки, нужно написать <code>\\{…\\}</code>. Если писать просто <code>{…}</code>, скобки играют роль не символов, а «объединителей» TeX-овских команд. Кстати, в роли объединителей ими не нужно злоупотреблять; используйте их только при необходимости.",
        "severity": 0
    },
    # todo
    "MANUAL_LISTS": {
        "msg": "Не следует вручную создавать нумерованные списки. Пишите так: <code>\\begin{enumerate}\\item Во-первых, \\item Во-вторых … \\end{enumerate}</code>. Подробнее об оформлении списков в \\(\\LaTeX\\) можно прочитать, например, <a href=\"https://www.overleaf.com/learn/latex/Lists\" target=\"_blank\">здесь</a>.",
        "severity": 0
    }
}

error_descriptions[SupportedLanguages.EN] = {
    "PARSE_ERROR": {
        "msg": "Error parsing LaTeX code. Possibly, a formula is not closed with a dollar sign somewhere, or conversely, a dollar is forgotten before a formula.",
        "severity": 0
    },
    "DOUBLE_DOLLARS": {
        "msg": "Double dollars for display math should be avoided in \\(\\LaTeX\\). Use <code>\\[</code>\u2026<code>\\]</code> for display style formulas. See <a href=\"https://tex.stackexchange.com/questions/503/why-is-preferable-to\" target=\"_blank\">explanation</a>.",
        "severity": 0
    },
    "CONSECUTIVE_DISPLAY_FORMULAE": {
        "msg": "Two consecutive display style formulas have been detected. This should be avoided. Use <code>aligned</code> environment or its analogues to properly typeset long formula that does not fit into single line. See e.g. <a href=\"https://www.overleaf.com/learn/latex/Aligning_equations_with_amsmath\" target=\"_blank\">these guidelines</a>.",
        "severity": 0
    },
    "LINEBREAK_AFTER_DISPLAY_FORMULAE": {
        "msg": "Line break detected immediately next to a display formula. After a display formula, the transition to a new line is automatic. If you wanted to increase the vertical spacing to the next text after the formula, use other means.",
        "severity": 0
    },
    "LINEBREAK_BEFORE_DISPLAY_FORMULAE": {
        "msg": "Line break detected immediately before a display formula. Before a display formula, the transition to a new line is automatic. If you wanted to increase the vertical spacing between the text and the formula.",
        "severity": 0
    },
    "EQNARRAY_USED": {
        "msg": "Avoid <code>eqnarray</code> environment (see <a href=\"https://tex.stackexchange.com/a/197\" target=\"_blank\">details</a>). Use e.g. <code>align</code> instead.",
        "severity": 0
    },
    "ABBREVIATIONS_WITH_SPACE": {
        "msg": "Abbreviations like <em>i.&thinsp;e.</em>, <em>e.&thinsp;g.</em>, <a href=\"https://ru.wikipedia.org/wiki/Q.E.D.\" target=\"_blank\"><em>q.&thinsp;e.&thinsp;d.</em></a> and similar <a href=\"https://new.gramota.ru/spravka/buro/search-answer?s=296030\">should be formatted with a space</a> (see also <a href=\"https://popravilam.com/blog/105-probel-v-sokrashcheniyah.html\" target=\"_blank\">here</a>), but there is a peculiarity: it is necessary to use a <em>non-breaking</em> space <code>~</code> or, even better, <a href=\"https://ru.wikipedia.org/wiki/%D0%A3%D0%B7%D0%BA%D0%B8%D0%B9_%D0%BF%D1%80%D0%BE%D0%B1%D0%B5%D0%BB\" target=\"_blank\"><em>thin space</em></a> <code>\\,</code> \u2014 this is a non-breaking space slightly narrower than the usual one. For example: <code>q.\\,e.\\,d.</code>. Otherwise, a case may occur during text wrap when part of the abbreviation remains on the line and part moves to the next. Thin space also looks better in this case than a regular space. It is also used when typing <a href=\"https://new.gramota.ru/spravka/letters/78-init\" target=\"_blank\">initials</a>, for example: <code>J.\\,D.~Smith</code> or <code>J.\\,D.\\,Smith</code>.",
        "severity": 0
    },
    "DASH_HYPHEN": {
        "msg": "Looks like you have wrongly used hyphen instead of dash. You can typeset M-dash \u201c\u2014\u201d with <code>---</code> and N-dash \u201c\u2013\u201d with <code>--</code> in \\(\\LaTeX\\). See <a href=\"https://www.grammarly.com/blog/hyphens-and-dashes/\" target=\"_blank\">details</a>.",
        "severity": 0
    },
    "DASH_IN_MATH_MODE": {
        "msg": "Looks like you wanted to typeset a hyphen. But when hyphen appears in math mode \\(\\LaTeX\\) treats it as minus. So you should make this hyphen part of the text that follows the formula, not the formula itself.",
        "severity": 0
    },
    "DASH_SURROUND_WITH_SPACES": {
        "msg": "An em dash should be surrounded by spaces on both sides. A particularly chic look is achieved by making one or both of these spaces non-breaking, so the em dash does not 'hang on the edge of the abyss' during line breaks. More about em dashes, hyphens, and similar can be found <a href=\"https://webstyle.sfu-kras.ru/tire-defis\" target=\"_blank\">here</a> and <a href=\"https://habrahabr.ru/post/20588/\" target=\"_blank\">here</a>. And, of course, it's never too late to read <a href=\"https://www.artlebedev.ru/kovodstvo/sections/97/\" target=\"_blank\">A. Lebedev</a>.",
        "severity": 0
    },
    "RU_ORDINAL_ABBREVIATION": {
        "msg": "Rules for abbreviating ordinal numbers are violated <a href=\"https://new.gramota.ru/spravka/letters/22-spravka/letters/87-rubric-99\" target=\"_blank\">here</a>.",
        "severity": 0
    },
    "PARAGRAPH_BREAK_BEFORE_DISPLAY_FORMULA": {
        "msg": "An empty line makes \\(\\LaTeX\\) to start a new paragraph even if this line happens just before a display-style formula. Most of the time it is not needed, as paragraphs should not start with formulae, and as display math already have enough space before and after.",
        "severity": 0
    },
    "UNNECESSARY_FORMULA_BREAK": {
        "msg": "Looks like some math needs concatenating. For instance, instead of <code>$x$, $y$ $\\in$ $A$</code> write <code>$x,\\,y\\in A$</code>; instead of <code>$a$ = $b+c$</code> write <code>$a=b+c$</code> etc.",
        "severity": 0
    },
    "CENTERING": {
        "msg": "Explicit centering should only be used in figure an table environments, and sometimes for headings. Avoid centering of non-display-style formulae.",
        "severity": 0
    },
    "LOW_LEVEL_FONT_COMMANDS": {
        "msg": "Instead of using low-level TeX commands <code>{\\it \u2026}</code>, <code>{\\bf \u2026}</code>. Instead use <code>\\textit{\u2026}</code>, <code>\\textbf{\u2026}</code> etc (see <a href=\"https://tex.stackexchange.com/questions/41681/correct-way-to-bold-italicize-text\" target=\"_blank\">details</a>.) It is also recommended to use <code>\\emph{\u2026}</code> command to emphasize portions of the text instead of italicizing, as it nicely works also inside italicized text blocks.",
        "severity": 0
    },
    "WRONG_QUOTES": {
        "msg": "Do not use \u201cprogrammers\u2019 quotes\u201d like <code>\"\u2026\"</code> to quote in \\(\\LaTeX\\). Use <code>``\u2026''</code> instead.",
        "severity": 0
    },
    "WRONG_SAME_QUOTES": {
        "msg": "Opening and closing quotes are usually different: avoid <code>''\u2026''</code> or <code>``\u2026``</code>, make it <code>``\u2026''</code> instead.",
        "severity": 0
    },
    "QUOTES_IN_MATH": {
        "msg": "Quote symbol detected in math mode. If you need to put a mathematical \u201cprime\u201d symbol use <code>\\prime\\prime</code> or just keyboard apostrophes <code>''</code>.",
        "severity": 0
    },
    "LATIN_LETTER_OUTSIDE_MATH_RU": {
        "msg": "Even one letter, if it has a mathematical meaning (if it is the name of a mathematical object) \u2014 should be typed in math mode.",
        "severity": 0
    },
    "MATH_SEMANTICS_OUTSIDE_MATH": {
        "msg": "Commands that typeset <em>mathematical</em> symbols should generally be placed inside math mode (in spite of some of them work in text mode also).",
        "severity": 0
    },
    "LATIN_LETTER_C_MISUSED": {
        "msg": "Looks like you have accidentally used latin c instead of cyrillic \u0441.",
        "severity": 0
    },
    "CYRILLIC_LETTER_C_MISUSED": {
        "msg": "Looks like you have accidentally used one of cyrillic letters a,e,c instead of their latin counterparts.",
        "severity": 0
    },
    "MULTIPLICATION_SIGN": {
        "msg": "Symbol <code>*</code> is used to denote multiplication in programming, not in mathematics. Use <code>\\cdot</code> or <code>\\times</code> instead (<code>\\times</code> is used only in special circumstances though).",
        "severity": 0
    },
    "SPACE_BEFORE_PUNCTUATION_MARK": {
        "msg": "There should be no space symbol inserted before colon, semicolon, comma, full stop, exclamation mark or question mark.",
        "severity": 0
    },
    "SPACE_BEFORE_PARENTHESIS": {
        "msg": "There should be a space before the opening parenthesis.",
        "severity": 0
    },
    "SPACE_AFTER_PUNCTUATION_MARK": {
        "msg": "There should be a space after colon, semicolon, comma, full stop, exclamation mark or question mark.",
        "severity": 0
    },
    "SPACE_AFTER_PARENTHESIS": {
        "msg": "There should be no space after the opening parenthesis.",
        "severity": 0
    },
    "CAPITALIZATION_AFTER_PUNCTUATION_MARK": {
        "msg": "Colon, semicolon, and comma do not mark the end of the sentence, and thus the word following them should generally not be capitalized.",
        "severity": 0
    },
    "CAPITALIZATION_AFTER_PERIOD": {
        "msg": "Sentences should start with capital letter.",
        "severity": 0
    },
    "PERIOD_BEFORE_NEXT_SENTENCE": {
        "msg": "Sentences should ultimately end with full stop even if they end with formula.",
        "severity": 0
    },
    "LEFT_RIGHT_RECOMMENDED": {
        "msg": "When in need of typesetting paired delimiters (parentheses, brackets, braces, absolute value symbol, norm symbol) around some large formula, it is recommended to add <code>\\left</code> and <code>\\right</code> commands. Compare how ugly is the PDF representation of a formula <code>\\[(\\frac{a}{b})^2\\]</code> and how beautiful is the proper variant of it: <code>\\[\\left(\\frac{a}{b}\\right)^2\\]</code>. On the other hand, <code>\\left\u2026 \u2026 \\right\u2026</code> should not be overused, use them just when you see the mismatch between the formula hight and the delimiter height. See <a href=\"https://tex.stackexchange.com/a/58641\" target=\"_blank\">details</a>. You can also employ <a href=\"https://tex.stackexchange.com/a/1765\" target=\"_blank\"><code>\\DeclarePairedDelimiter</code></a> command from <em>mathtools</em> library.",
        "severity": 10
    },
    "SYMBOLIC_LINKS": {
        "msg": "Employ symbolic cross-referencing, e.g.: <code>Note that \\begin{equation}\\label{eqSquares} a^2-b^2=(a-b)(a+b). \\end{equation} From~\\eqref{eqSquares} it follows that\u2026</code>. See <a href=\"https://www.overleaf.com/learn/latex/Cross_referencing_sections%2C_equations_and_floats\" target=\"_blank\">details</a>.",
        "severity": 0
    },
    "EQREF_INSTEAD_OF_REF": {
        "msg": "Write <code>\\eqref{\u2026}</code> instead of <code>(\\ref{\u2026})</code>.",
        "severity": 0
    },
    "NONBREAKABLE_SPACE_BEFORE_REF": {
        "msg": "Commands <code>\\ref</code>, <code>\\eqref</code> and such are recommended to be prepended with unbreakable space <code>~</code>, e.g.: <code>\u2026Using~\\eqref{eqMaxwell} we obtain\u2026</code>.",
        "severity": 0
    },
    "ELLIPSIS_LDOTS": {
        "msg": "You should use <code>\\ldots</code> for ellipsis in math mode instead of <code>...</code>. If you use <em>amsmath</em> package you can use <code>\\dotsc</code> command for ellipsis in typical mathematical enumerations.",
        "severity": 0
    },
    "TRIVIAL_LABEL": {
        "msg": "It is definitely a good idea to use symbolic cross-references instead of hard-coding explicit numbers, but you should avoid using non-semantic names of the references like <code>\\label{eq1}</code>. Just as a variable with the name <code>var1</code> is frowned upon in software development. Take time and think on some meaningful name, e.g. <code>eqCauchy</code>, <code>eqBinomialSymmetry</code>, <code>eqMain</code> (the latter if fine if this is indeed the most important formula in your proof) etc.",
        "severity": 5
    },
    "REPLACE_MBOX_WITH_TEXT": {
        "msg": "To place text inside a formula avoid using low-level <code>\\mbox</code> and <code>\\hbox</code> commands. Use <code>\\text</code> instead. See <a href=\"https://tex.stackexchange.com/questions/70632/difference-between-various-methods-for-producing-text-in-math-mode\" target=\"_blank\">details</a>.",
        "severity": 0
    },
    "TEXT_IN_MATH_MODE": {
        "msg": "To insert text inside a math formula use <code>\\text</code> command, e.g.: <code>Consider a set $A_{\\text{good}}$</code>.",
        "severity": 0
    },
    "INCORPORATE_NOT": {
        "msg": "Although <code>\\not</code> command does work to get a negated (crossed-out) sign, the <a href=\"https://tex.stackexchange.com/a/141011\" target=\"_blank\">better way</a> is to search for the dedicated command. For instance, <code>\\ne</code> is better than <code>\\not=</code>, <code>\\notin</code> is better than <code>\\not\\in</code> etc.",
        "severity": 0
    },
    "OVER_VS_FRAC": {
        "msg": "The command <code>\\over</code> is a low-level TeX command and <a href=\"https://tex.stackexchange.com/a/73825\" target=\"_blank\">should be avoided</a> in \\(\\LaTeX\\). Replace <code>A \\over B</code> with <code>\\frac{A}{B}</code>.",
        "severity": 0
    },
    "CHOOSE_VS_BINOM": {
        "msg": "The command <code>\\choose</code> is a low-level TeX command and <a href=\"https://tex.stackexchange.com/a/127711\" target=\"_blank\">should be avoided</a> in \\(\\LaTeX\\). Replace <code>A \\choose B</code> with <code>\\binom{A}{B}</code>.",
        "severity": 0
    },
    "SETS_IN_BBFONT": {
        "msg": "Standard number sets should be typeset in \u201cblackboard bold\u201d font: e.g. the usages of <code>N</code> as a set of naturals should be replaced with <code>\\mathbb{N}</code>.",
        "severity": 0
    },
    "MOD_NOT_A_COMMAND": {
        "msg": "Use <code>\\bmod</code> or <code>\\pmod</code> to typeset \\(\\bmod\\) with roman font.",
        "severity": 0
    },
    "TILDE_INEFFECTIVE_AS_NBSP": {
        "msg": "The symbol <code>~</code> denotes the unbreakable space in \\(\\LaTeX\\). It should not be surrounded with spaces: you loose unbreakability. For instance, instead of <code>Formula ~\\eqref{\u2026} implies\u2026</code> you should write <code>Formula~\\eqref{\u2026} implies\u2026</code>",
        "severity": 0
    },
    "INDENTATION_WITH_SPACES": {
        "msg": "Avoid using multiple spaces to indent text. Use proper \\(\\LaTeX\\) spacing command instead, see <a href=\"https://tex.stackexchange.com/a/74354\" target=\"_blank\">details</a>.",
        "severity": 0
    },
    "LE_AS_SINGLE_COMMAND": {
        "msg": "To typeset \\(\\le\\) or \\(\\ge\\) employ <code>\\le</code> and <code>\\ge</code> commands respectively. The arrow \\(\\Leftarrow\\) is typeset with <code>\\Leftarrow</code> command. Software developing combinations like <code><=</code> and <code>>=</code> are of no use in \\(\\LaTeX\\).",
        "severity": 0
    },
    "PUNCTUATION_AFTER_DISPLAY_MATH": {
        "msg": "When you place a punctuation mark right after a display-style formula, it is going to be thorn away from it and placed on a separate line (just look at the compiled PDF). So if you need e.g. to place a comma right after display formula, you should make this comma part of the formula itself.",
        "severity": 0
    },
    "BACKSLASH_NEEDED": {
        "msg": "Words \\(\\min\\), \\(\\max\\) and such in mathematical formulas are operator names and should be typeset with roman font. There are corresponding <em>commands</em> of \\(\\LaTeX\\) <code>\\min</code>, <code>\\max</code>, <code>\\lim</code>, <code>\\deg</code>, and others, that will do that work for you. See the list of these commands <a href=\"https://www.overleaf.com/learn/latex/Operators#Reference_guide\" target=\"_blank\">here</a>. If you cannot see the command that you need, write something like <code>\\operatorname{min}</code> or, even better, <a href=\"https://tex.stackexchange.com/a/67529\" target=\"_blank\">declare your own operator</a> with <code>\\DeclareMathOperator</code> command in the document preamble.",
        "severity": 0
    },
    "CDOT_FOR_READABILITY": {
        "msg": "When typesetting products of numbers, fractions, binomial coefficients and such, consider using explicit multiplication dot <code>\\cdot</code> occasionally.",
        "severity": 5
    },
    "GRAPHICS_IN_MATH_MODE": {
        "msg": "The <code>\\includegraphics</code> command should not be used in math mode except for you really know what you are doing with it. For instance, to center a figure on screen, use <code>center</code> and <code>figure</code> environments instead of surrounding your <code>\\includegraphics</code> with display math delimiters.",
        "severity": 0
    },
    "UNNECESSARY_MATH_MODE": {
        "msg": "If the only content of a formula is a single symbols which is neither a digit nor a single letter \u2014 that is suspicious! Most likely you either did not need math mode here or unnecessarily broke a formula into pieces.",
        "severity": 0
    },
    "NO_SPACE_AFTER_COMMAND_BEFORE_CYRILLIC": {
        "msg": "",
        "severity": 0
    },
    "TEXT_COMMANDS_IN_MATH_MODE": {
        "msg": "Text style commands <code>\\textbf</code>, <code>\\textit</code> and such should be avoided in math mode. There are math-mode commands for some font styles. For instance, <code>\\mathbf</code> for mathematical bold font.",
        "severity": 0
    },
    "LIMITS_UNNECESSARY_IN_DISPLAY_MODE": {
        "msg": "The <code>\\limits</code> command can be frequently omitted in display formulas: the operator limits are typically automatically placed above and below the operator even without it in display math mode.",
        "severity": 0
    },
    "FORMULA_NEIGHBOURING_REFERENCE": {
        "msg": "A formula that directly neighbours a reference typically looks inferior. You can always think on some word to insert in between them.",
        "severity": 0
    },
    "UNICODE_SQRT": {
        "msg": "To typeset the square root symbol use <code>\\sqrt</code> instead of UNICODE symbol. As a benefit, you will also get a beautiful stretching overline.",
        "severity": 0
    },
    "NUMBERED_MATH_NEEDS_REFERENCING": {
        "msg": "If a formula has a visible number, then this number must be used somewhere in the text to refer to the formula. And if it is not necessary to refer to the formula, then it should not be numbered.",
        "severity": 0
    },
    "NO_CONCLUSION": {
        "msg": "It is rarely a good idea to end the proof or a problem solution right on some formula or a figure. It is best to add some concluding remark in the end, e.g.: <code>This concludes the proof.</code> or <code>Thus we finally get unknown value: $42$.</code> etc.",
        "severity": 5
    },
    "SUGGESTED_NEW_PARAGRAPH": {
        "msg": "To place a formula on a new line you should make that formula to be display-style. To start a new paragraph use <code>\\par</code> in text mode. Avoid using low-level line break <code>\\\\</code> except for <code>tabular</code>, <code>matrix</code>, or similar environments.",
        "severity": 0
    },
    "NUMERALS_AS_WORDS": {
        "msg": "Stylistically it usually looks better when small numerals are kept as words, not numbers. E.g. <em>Consider 2 cases</em> looks inferior to <em>Consider two cases</em>.",
        "severity": 10
    },
    "MID_IN_SET_COMPREHENSION": {
        "msg": "When describing sets the vertical line is placed using <code>\\mid</code> command. E.g. <code>\\{x^2\\mid x\\in\\mathbb{Z}\\}</code>. In other contexts you should generally avoid <code>\\mid</code>. For instance, for absolute value use <code>|x|</code> or <code>\\lvert x \\rvert</code>.",
        "severity": 0
    },
    "FLOOR_FUNCTION_NOTATION": {
        "msg": "In modern mathematical literature the integer part of number \\(x\\) is denoted as <code>\\lfloor x\\rfloor</code>, not <code>[x]</code>.",
        "severity": 0
    },
    "ITALIC_INSTEAD_OF_EMPH": {
        "msg": "Consider using <code>\\emph{\u2026}</code> for text emphasis instead of italicizing, as it nicely works also inside italicized text blocks.",
        "severity": 0
    },
    "PARAGRAPH_STARTS_WITH_FORMULA": {
        "msg": "A paragraph should not start with a formula. If you need to place a formula on a separate line, typeset it in display mode using <code>\\[\u2026\\]</code>.",
        "severity": 0
    },
    "SENTENCE_STARTS_WITH_FORMULA": {
        "msg": "Typically a sentence should not start with a formula. You can pretty much always add some introductory words. For instance, instead of <code>$G$ is connected.</code> write <code>Graph $G$ is connected.</code>",
        "severity": 0
    },
    "SENTENCE_STARTS_WITH_NUMBER": {
        "msg": "Typically a sentence should not start with a number. Consider using numerals or reordering the sentenced. <code>5 persons can form a queue in $5!$ ways.</code> you could write <code>Five persons\u2026</code> or <code>Note that 5 persons\u2026</code>",
        "severity": 5
    },
    "RUSSIAN_TYPOGRAPHY_PECULIARITIES": {
        "msg": "Strictly speaking, this is not an error, but in the Russian typographic tradition, it is customary to denote the empty set with the symbol <code>\\varnothing</code>, not <code>\\emptyset</code> (the latter is more flattened). Similarly, instead of <code>\\epsilon</code> it is better to write <code>\\varepsilon</code>, and instead of <code>\\phi</code> write <code>\\varphi</code>.",
        "severity": 0
    },
    "BETTER_TO_USE_WORDS_THEN_MATH": {
        "msg": "Constructs like <code>the number of elements $=m^2$</code> are unacceptable in proper mathematical writing except for taking short personal notes. Symbols  \\(=, \\gt, \\geqslant\\) and such should be replaced with words if surrounded by words: <code>\u2026does not exceed $m^2$</code>, <code>\u2026equals $m^2$</code> etc.",
        "severity": 0
    },
    "MATH_ENVIRONMENT_VERBOSITY_WARNING": {
        "msg": "The <code>math</code> environment is not popular in \\(\\LaTeX\\) community. You can use a shorter (and absolutely equivalent in terms of what you get of it) construct <code>\\(\u2026\\)</code> or <code>$\u2026$</code>. Note also that <em>math</em> does not make a display-style formula, it opens the standard inline-math mode. For display math use <code>\\[\u2026\\]</code> or one of the environments <em>equation</em>, <em>array</em> etc.",
        "severity": 0
    },
    "USE_DIVIDES_INSTEAD_OF_VDOTS": {
        "msg": "Command <code>\\vdots</code> should be primarily used for typesetting matrices as a vertical ellipsis. When writing on divisibility of numbers, consider using vertical line (\u201c\u2026devides\u2026\u201d) instead of triple vertical dots (\u201c\u2026is divisible by\u2026\u201d).",
        "severity": 0
    },
    "MAKE_LONG_FORMULA_DISPLAY": {
        "msg": "Long formulas that take a lot of screen space should generally be thoughtfully typeset with proper AMS environments: see the list of these environments and their usecases <a href=\"https://www.overleaf.com/learn/latex/Aligning_equations_with_amsmath\" target=\"_blank\">here</a>.",
        "severity": 0
    },
    "EN_ORDINAL_ABBREVIATION": {
        "msg": "Possibly wrong ordinal abbreviation, see <a href=\"https://www.grammarly.com/blog/how-to-write-ordinal-numbers-correctly/\" target=\"_blank\">details</a>.",
        "severity": 0
    },
    "LATE_DEFINITION": {
        "msg": "Instead of writing <q><code>$x=a+b$</code>, where <code>$a=\u2026</code></q> it's better to introduce all the letters first and then write the expression containing these letters. See <a href=\"https://1drv.ms/w/s!AiAwrmxQ9QLrjOtTQqAXWdl3ryK5Jg?e=vEMU6W\" target=\"_blank\">P. Halmos's article</a>, the section 'Use Words Correctly'.",
        "severity": 0
    },
    "INVISIBLE_BRACES": {
        "msg": "To typeset braces you should write <code>\\{\u2026\\}</code>. Without backslashes the braces <code>{\u2026}</code> are the semantic delimiters in code, but are not displayed in the PDF.",
        "severity": 0
    },
    "MANUAL_LISTS": {
        "msg": "Avoid manual numbering in lists. Employ <em>enumerate</em> environment: <code>\\begin{enumerate}\\item Firstly\u2026 \\item Secondly\u2026 \\end{enumerate}</code>. Learn more about typesetting lists <a href=\"https://www.overleaf.com/learn/latex/Lists\" target=\"_blank\">here</a>.",
        "severity": 0
    },
    "MISMATCHED_MATH_DELIMITERS": {
        "msg": "Possibly mismatched math delimiters.",
        "severity": 0
    },
    "EN_ORDINAL_ABBREVIATION_IN_MATH": {
        "msg": "Ordinal abbreviations are textual, not mathematical pieces. See <a href=\"https://tex.stackexchange.com/a/4119\" target=\"_blank\">details</a> on how to properly typeset them in \\(\\LaTeX\\) if you\u2019d like to keep the superscript style.",
        "severity": 0
    },
    "LATIN_LETTER_OUTSIDE_MATH_EN": {
        "msg": "Even a single letter (if this letter has a mathematical semantics) should be typeset in math mode.",
        "severity": 0
    }
}


from setuptools import setup, find_packages

setup(
    name='latexcheck',
    version='1.2.5',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    entry_points={
        'console_scripts': ['latexcheck=latexcheck:main']
    },
    install_requires=[
        'texsoup>=0.3.1'
    ],
    python_requires='>=3.8',
    author='Alex Dainiak',
    author_email='dainiak@gmail.com',
    description='LaTeX linter',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='http://github.com/dainiak/latexcheck-py/',
    classifiers=[
        'License :: OSI Approved :: MIT License'
    ],
)